#include <windows.h>
#include <stdio.h>
#include <vector>
#include <algorithm>
using namespace std;

#define MSG_POPUP WM_USER+100
#define MNU_EXIT 1
#define MNU_AUTOSTART 2
#define MNU_NONSTD_RATIO 3
#define MNU_LORES 4
#define MNU_SAVE_SETTINGS 5
#define MNU_RESOLUTIONS 6
#define DEFAULT_MIN_HORZ_RES 640

HANDLE hMutex;
HINSTANCE hInst;
const char szAppName[]="Resolution Changer";

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

class Settings {
public:
	static unsigned MinHorzRes;
	static bool DisplayNonStdRatio;
	static bool AutoStart;

	static void reset() {
		MinHorzRes = DEFAULT_MIN_HORZ_RES;
		DisplayNonStdRatio = false;
		AutoStart = false;
	}

	static void loadFromRegistry() {
		HKEY key;
		unsigned long type, size;

		reset();
		if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\ChangeRes", 0, 
			KEY_READ, &key) == ERROR_SUCCESS) 
		{ 
			DWORD buffer;
			size = 4;
			if (RegQueryValueEx(key, "DisplayLoRes", 0, &type, (LPBYTE)&buffer, &size) == ERROR_SUCCESS)
				if (type==REG_DWORD && size==4)
					MinHorzRes = buffer ? 0 : DEFAULT_MIN_HORZ_RES;
			size = 4;
			if (RegQueryValueEx(key, "DisplayNonStdRatio", 0, &type, (LPBYTE)&buffer, &size) == ERROR_SUCCESS)
				if (type==REG_DWORD && size==4)
					DisplayNonStdRatio = buffer!=0;
			RegCloseKey(key);
		}

		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Run",
			0, KEY_READ, &key) == ERROR_SUCCESS) 
		{
			char buffer[256];
			size = sizeof(buffer);
			if (RegQueryValueEx(key, "ChangeRes", 0, &type, (LPBYTE)&buffer, &size) == ERROR_SUCCESS)
				if (type==REG_SZ)
					AutoStart = true;
			RegCloseKey(key);
		}
	}

	// all=true: minden belltst ment
	//    false: csak az autosave-et
	static bool saveToRegistry(bool all) {
		HKEY key;
		DWORD disp;
		bool ok=true;

		// menti a program belltsait a HKEY_CURRENT_USER/Software/ChangeRes gra
		if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\ChangeRes", 0, "", 0, KEY_WRITE, NULL, &key, &disp) == ERROR_SUCCESS) {
			DWORD buffer;
			if (all) {
				buffer=(DWORD)DisplayNonStdRatio;
				if (RegSetValueEx(key, "DisplayNonStdRatio", 0, REG_DWORD, (LPBYTE)&buffer, 4) != ERROR_SUCCESS) ok=false;
				buffer=(unsigned long)(MinHorzRes==0);
				if (RegSetValueEx(key, "DisplayLoRes", 0, REG_DWORD, (LPBYTE)&buffer, 4) != ERROR_SUCCESS) ok=false;
			}
		}

		// fizikailag ki/bekapcsolja az automatikus indtst
		if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Run", 0, "", 0, KEY_ALL_ACCESS, NULL, &key, &disp) == ERROR_SUCCESS) {
			if (AutoStart) {
				const int buflen=260;
				TCHAR buffer[buflen];
				GetModuleFileName(0, buffer, buflen-1);
				if (RegSetValueEx(key, "ChangeRes", 0, REG_SZ, (LPBYTE)buffer, (DWORD)strlen(buffer)) != ERROR_SUCCESS) ok=false;
			} else RegDeleteValue(key, "ChangeRes");
			RegCloseKey(key);
		} else ok=false;

		return ok;
	}
};

unsigned Settings::MinHorzRes;
bool Settings::DisplayNonStdRatio;
bool Settings::AutoStart;

class DisplayMode {
	friend struct DisplayMode_Less;
	unsigned width, height, frequency, bitsPerPixel;

	int isValidRange() {
		return 
			width>=320 && width<=2048 &&
			height>=200 && height<=2048 &&
			(bitsPerPixel==0 || bitsPerPixel==1 || bitsPerPixel==2 ||
			bitsPerPixel==4 || bitsPerPixel==8 || bitsPerPixel==15 ||
			bitsPerPixel==16 || bitsPerPixel==24 || bitsPerPixel==32) &&
			frequency<=200 ? SUCCESS : RANGE_ERROR;
	}
	
public:
	static const int 
		SUCCESS = 0,
        PARSE_ERROR = 1,
		RANGE_ERROR = 2;

	DisplayMode() { }

	DisplayMode(DEVMODE devm) {
		width = devm.dmPelsWidth;
		height = devm.dmPelsHeight;
		frequency = devm.dmDisplayFrequency;
		bitsPerPixel = devm.dmBitsPerPel;
	}

	void setSafe() {
		width=640;
		height=480;
		bitsPerPixel=32;
		frequency=60;
	}

	int parseString(char *s) {
		char d1, d2;
		int n = sscanf(s, "%dx%d%c%d%c%d",
			&width, &height, &d1, &bitsPerPixel, &d2, &frequency);
		if (n==6 && d1=='x' && d2=='@') return isValidRange();
		if (n==4 && d1=='x') { 
			frequency = 0; 
			return isValidRange(); 
		}
		if (n==4 && d1=='@') {
			frequency = bitsPerPixel;
			bitsPerPixel = 0;
			return isValidRange();
		}
		if (n==2) { 
			bitsPerPixel = 0;
			frequency = 0; 
			return isValidRange(); 
		}
		return PARSE_ERROR;
	}

	DisplayMode& operator= (const DisplayMode &m) {
		width = m.width;
		height = m.height;
		frequency = m.frequency;
		bitsPerPixel = m.bitsPerPixel;
		return *this;
	}

	bool operator< (const DisplayMode &m) const {
		return width==m.width && height==m.height &&
			bitsPerPixel==m.bitsPerPixel && frequency<m.frequency;
	}

	bool operator== (const DisplayMode &m) const {
		return width==m.width && height==m.height &&
			bitsPerPixel==m.bitsPerPixel;
	}

	unsigned getBPP() { return bitsPerPixel; }

	void toString(bool bpp, char* s) {
		if (bpp)
			sprintf(s, "%dx%dx%d %d Hz", width, height, bitsPerPixel, frequency);
		else
			sprintf(s, "%dx%d %d Hz", width, height, frequency);
	}

	void findMaxFreq() {
		DEVMODE devm;
		int i=0;
		frequency=0;
		while (EnumDisplaySettings(NULL, i++, &devm)) {
			DisplayMode m(devm);
			if (*this<m) frequency = devm.dmDisplayFrequency;
		}
	}

	long apply() {
		if (!bitsPerPixel) {
			DisplayMode cur;
			cur.getCurrentMode();
			bitsPerPixel = cur.bitsPerPixel;
		}
		if (!frequency) findMaxFreq();

		DEVMODE devm;
		ZeroMemory(&devm, sizeof(DEVMODE));
		devm.dmSize = sizeof(DEVMODE);
		devm.dmPelsWidth = width;
		devm.dmPelsHeight = height;
		devm.dmBitsPerPel = bitsPerPixel;
		devm.dmDisplayFrequency = frequency;
		devm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT |
			DM_DISPLAYFREQUENCY | DM_BITSPERPEL;
		long result = ChangeDisplaySettings(&devm, CDS_UPDATEREGISTRY);
		if (result==DISP_CHANGE_NOTUPDATED)
			result = ChangeDisplaySettings(&devm, 0);
		return result;
	}

	void getCurrentMode() {
		HDC hdcScreen = CreateDC("DISPLAY", NULL, NULL, NULL);
		width = GetDeviceCaps(hdcScreen, HORZRES);
		height = GetDeviceCaps(hdcScreen, VERTRES);
		frequency = GetDeviceCaps(hdcScreen, VREFRESH);
		bitsPerPixel = GetDeviceCaps(hdcScreen, BITSPIXEL);
		DeleteDC(hdcScreen);
	}
};

struct DisplayMode_Less {
	bool operator() (const DisplayMode &n, const DisplayMode &m) {
		return n.width==m.width && n.height==m.height &&
			n.bitsPerPixel==m.bitsPerPixel && n.frequency<m.frequency;
    }
};

class DisplayModeList {
private:
	vector<DisplayMode> dm;

	void clear() {
		dm.clear();
	}

	void add(DEVMODE devm) {
		DisplayMode m(devm);
		vector<DisplayMode>::iterator i;
		for (i=dm.begin(); i!=dm.end(); i++)
			if (*i<m) { *i=m; return; }
		dm.push_back(m);
	}

	void sort() {
		std::sort(dm.begin(), dm.end(), DisplayMode_Less());
	}

public:
	int indexOf(DisplayMode m) const {
		vector<DisplayMode>::const_iterator i;
		for (i=dm.begin(); i!=dm.end() && !(*i==m); i++) ;
		if (i==dm.end()) return -1;
		else return (int)(i-dm.begin());
	}

	int count() const {
		return (int)dm.size();
	}

	void toString(int index, bool bpp, char *s) {
		dm[index].toString(bpp, s);
	}

	void getModes() {
		clear();
		DisplayMode cur;
		cur.getCurrentMode();
		unsigned BPP=cur.getBPP();
		DEVMODE devm;
		int i=0;
		while (EnumDisplaySettings(NULL, i++, &devm)) {
			if (devm.dmBitsPerPel == BPP && 
				devm.dmPelsWidth >= Settings::MinHorzRes &&
				(devm.dmPelsWidth*3 == devm.dmPelsHeight*4 ||
					Settings::DisplayNonStdRatio))
				add(devm);
		}
		sort();
	}

	void apply(int index) {
		dm[index].apply();
	}
};

/*===========================================================================*/
/* Command line conversion routines                                          */
/* http://www.koders.com/c/fidE7CF9B9B368FDB967F89CC6C228C3C7BC69691C2.aspx  */
/*===========================================================================*/
class CmdLine {
	// Returns the first character in the next argument
	static char *winDrvCmdLineGetNextFirst(char *lpCmdLine) {
		while (*lpCmdLine == ' ' && *lpCmdLine != '\0')
			lpCmdLine++;
		return (*lpCmdLine == '\0') ? NULL : lpCmdLine;
	}

	// Returns the first character after the next argument
	static char *winDrvCmdLineGetNextEnd(char *lpCmdLine) {
		int InString = FALSE;
		while (((*lpCmdLine != ' ') && (*lpCmdLine != '\0')) ||
			(InString && (*lpCmdLine != '\0'))) {
				if (*lpCmdLine == '\"')
					InString = !InString;
				lpCmdLine++;
			}
			return lpCmdLine;
	}

	// Returns an argv vector and takes argc as a pointer parameter
	// Must free memory argv on exit
	static char **winDrvCmdLineMakeArgv(char *lpCmdLine, int *argc) {
		int elements = 0, i;
		char *tmp;
		char **argv;
		char *argstart, *argend;

		tmp = winDrvCmdLineGetNextFirst(lpCmdLine);
		if (tmp != 0) {
			while ((tmp = winDrvCmdLineGetNextFirst(tmp)) != NULL) {
				tmp = winDrvCmdLineGetNextEnd(tmp);
				elements++;
			}
		}
		argv = (char **) malloc(4*(elements + 2));
		argv[0] = "";
		argend = lpCmdLine;
		for (i = 1; i <= elements; i++) {
			argstart = winDrvCmdLineGetNextFirst(argend);
			argend = winDrvCmdLineGetNextEnd(argstart);
			if (*argstart == '\"')
				argstart++;
			if (*(argend - 1) == '\"')
				argend--;
			*argend++ = '\0';
			argv[i] = argstart;
		}
		argv[elements + 1] = NULL;
		*argc = elements + 1;
		return argv;
	}

public:
	static int analyzeArgs(char *szCmdLine, DisplayMode &dm) {
		char *cmdline = (char*)malloc(strlen(szCmdLine)+1);
		char **argv;
		char errormsg[300];
		int argc;
		strcpy(cmdline, szCmdLine);
		argv = winDrvCmdLineMakeArgv(cmdline, &argc);

		int start=0;  // 0: menu  1: exit  2: change mode then exit
		if (!start && argc>2) {
			MessageBox(NULL, "Too many parameters", "Error", MB_OK);
			start=1;  // exit after showing the dialog box
		}

		if (argc==2) {
			if (!start && (!strcmp(argv[1], "-?") || !strcmp(argv[1], "/?"))) {
				MessageBox(NULL,
					"changeres -safe\n"
					"      sets the display mode to 640x480x(32 bits/pixel) at 60 Hz\n"
					"changeres <width>x<height>[x<bpp>][@<freq>]\n"
					"      sets a custom display mode\n"
					//"changeres -list\n"
					//"      lists the legal display modes according to the monitor driver\n"
					, "Usage", MB_OK);
				start=1;  // exit after showing the dialog box
			}

			if (!start && !stricmp(argv[1], "-safe")) {
				dm.setSafe();
				start=2;
			}

			if (!start) {
				switch (dm.parseString(argv[1])) {
					case DisplayMode::PARSE_ERROR:
						sprintf(errormsg, "Invalid parameter: %s", argv[1]);
						MessageBox(NULL, errormsg, "Error", MB_OK);
						start = 1;
						break;
					case DisplayMode::RANGE_ERROR:
						sprintf(errormsg, "Range error in parameter: %s", argv[1]);
						MessageBox(NULL, errormsg, "Error", MB_OK);
						start = 1;
						break;
					default:  // success
						start = 2;
				}
			}
		}

		free(cmdline);
		free(argv);
		return start;
	}
};

DisplayModeList dm;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow) {
	MSG msg;
	WNDCLASSEX wndclass;

	DisplayMode m;
	int start=CmdLine::analyzeArgs(szCmdLine, m);

	if (start==0) {
		// fut-e mr az alkalmazs egy pldnyban?
		hMutex = CreateMutex(NULL, true, "ChangeRes");
		if (GetLastError()==ERROR_ALREADY_EXISTS) return 1;
		Settings::loadFromRegistry();

		hInst=hInstance;

		wndclass.cbSize        = sizeof(WNDCLASSEX);
		wndclass.style         = CS_HREDRAW | CS_VREDRAW;
		wndclass.lpfnWndProc   = WndProc;
		wndclass.cbClsExtra    = 0;
		wndclass.cbWndExtra    = 0;
		wndclass.hInstance     = hInstance;
		wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
		wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
		wndclass.lpszMenuName  = NULL;
		wndclass.lpszClassName = szAppName;
		wndclass.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

		RegisterClassEx(&wndclass);

		CreateWindow(szAppName, NULL, WS_OVERLAPPEDWINDOW, 0, 0, 0, 0,
			NULL, NULL, hInstance, NULL);

		while (GetMessage(&msg, NULL, 0, 0))
			DispatchMessage(&msg);
		return (int)msg.wParam;
	}

	if (start==2) {
		long result = m.apply();
		if (result != DISP_CHANGE_SUCCESSFUL &&
			result != DISP_CHANGE_RESTART)
			MessageBox(NULL, "Display mode change failed", "Error", MB_OK);
	}

	return 0;
}

HMENU BuildMenu() {
	int i;
	HMENU popupMenu, settingsMenu;
	popupMenu = CreatePopupMenu();
	settingsMenu = CreateMenu();

	// Settings men belseje
	AppendMenu(settingsMenu, 0, MNU_LORES, "Display low resolutions");
	if (Settings::MinHorzRes < DEFAULT_MIN_HORZ_RES)
		CheckMenuItem(settingsMenu, MNU_LORES, MF_CHECKED);
	AppendMenu(settingsMenu, 0, MNU_NONSTD_RATIO, "Display not 4/3 resolutions");
	if (Settings::DisplayNonStdRatio)
		CheckMenuItem(settingsMenu, MNU_NONSTD_RATIO, MF_CHECKED);
	AppendMenu(settingsMenu, 0, MNU_AUTOSTART, "Autostart");
	if (Settings::AutoStart)
		CheckMenuItem(settingsMenu, MNU_AUTOSTART, MF_CHECKED);
	AppendMenu(settingsMenu, MF_SEPARATOR, 0, NULL);
	AppendMenu(settingsMenu, 0, MNU_SAVE_SETTINGS, "Save settings");

	// Fmen
	AppendMenu(popupMenu, MF_POPUP, (UINT_PTR)settingsMenu, "Settings");
	AppendMenu(popupMenu, MF_SEPARATOR, 0, NULL);
	char resName[32];
	for (i=0; i<dm.count(); i++) {
		dm.toString(i, false, resName);
		AppendMenu(popupMenu, 0, MNU_RESOLUTIONS+i, resName);
	}
	DisplayMode cur;
	cur.getCurrentMode();
	if ((i = dm.indexOf(cur)) >= 0) 
		CheckMenuItem(popupMenu, MNU_RESOLUTIONS+i, MF_CHECKED);
	AppendMenu(popupMenu, MF_SEPARATOR, 0, NULL);
	AppendMenu(popupMenu, 0, MNU_EXIT, "Exit");

	return popupMenu;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
	POINT cpos;
	HMENU popup;
	static NOTIFYICONDATA nid;

	switch (iMsg) {
	case WM_CREATE:
		dm.getModes();
		nid.cbSize=sizeof(NOTIFYICONDATA);
		strcpy(nid.szTip, szAppName);
		ExtractIconEx("ChangeRes.ico", 0, NULL, &nid.hIcon, 1);
		nid.hWnd=hwnd;
		nid.uID=1;
		nid.uFlags = NIF_MESSAGE + NIF_ICON + NIF_TIP;
		nid.uCallbackMessage = MSG_POPUP;
		Shell_NotifyIcon(NIM_ADD, &nid);
		DestroyIcon(nid.hIcon);
		return 0;

	case MSG_POPUP:
		if (lParam==WM_LBUTTONDOWN || lParam==WM_RBUTTONDOWN) {
			GetCursorPos(&cpos);
			popup=BuildMenu();
			// tnjn el a men, ha mell kattintunk
			SetForegroundWindow(hwnd);
			TrackPopupMenu(popup, 0, cpos.x, cpos.y, 0, hwnd, 0);
			PostMessage(hwnd, WM_NULL, 0, 0);
			DestroyMenu(popup);
		}
		return 0;

	case WM_COMMAND:
		switch (LOWORD(wParam)) {
			case MNU_EXIT:
				SendMessage(hwnd, WM_CLOSE, 0, 0);
				break;
			case MNU_LORES:
				Settings::MinHorzRes = 
					Settings::MinHorzRes ? 0 : DEFAULT_MIN_HORZ_RES;
				dm.getModes();
				break;
			case MNU_NONSTD_RATIO:
				Settings::DisplayNonStdRatio =
					!Settings::DisplayNonStdRatio;
				dm.getModes();
				break;
			case MNU_AUTOSTART:
				Settings::AutoStart=!Settings::AutoStart;
				Settings::saveToRegistry(false);
				break;
			case MNU_SAVE_SETTINGS:
				Settings::saveToRegistry(true);
				break;
			default:
				if (LOWORD(wParam) >= MNU_RESOLUTIONS &&
					LOWORD(wParam) < MNU_RESOLUTIONS+dm.count())
						dm.apply(LOWORD(wParam)-MNU_RESOLUTIONS);
		}
		return 0;

	case WM_DESTROY:
		if (hMutex) ReleaseMutex(hMutex);
		Shell_NotifyIcon(NIM_DELETE, &nid);
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}
